using System;
using System.Collections.Generic;
using System.Query;
using QueryApp.DataWebService;
using System.Expressions;
using System.Collections;
using System.Reflection;

namespace QueryApplication {
	public class DataWebServiceQuery : IOrderedQueryable<CountryInfo> {
		public DataWebServiceQuery(Service dataWebService) {
			this.dataWebService = dataWebService;
		}

		Service dataWebService;
		Expression expression;

		#region Important methods
		#region IQueryable<CountryInfo.CreateQuery<S>
		IQueryable<S> IQueryable<CountryInfo>.CreateQuery<S>(Expression expression) {
			this.expression = expression;
			return (IQueryable<S>) this;
		}
		#endregion
		#region IQueryable.Expression
		Expression IQueryable.Expression {
			get {
				return expression ?? Expression.Constant(this);
			}
		}
		#endregion
		#region IEnumerable<CountryInfo>.GetEnumerator

		IEnumerator<CountryInfo> IEnumerable<CountryInfo>.GetEnumerator( ) {
			SearchInfo searchInfo = ParseExpression(expression);
			CountryInfo[] result = dataWebService.GetFilteredAndSortedData(searchInfo.FilterPrefix, searchInfo.SortOrder);
			return ((IEnumerable<CountryInfo>) result).GetEnumerator( );
		}

		#endregion
		#endregion

		// My own method, to interpret the expression
		#region ParseExpression
		SearchInfo ParseExpression(Expression expression) {
			SearchInfo result = new SearchInfo( );

			if (expression.NodeType != ExpressionType.MethodCall &&
				expression.NodeType != ExpressionType.MethodCallVirtual)
				throw new NotSupportedException("Only MethodCall(Virtual) expressions supported.");

			MethodCallExpression orderbyMethodCall = (MethodCallExpression) expression;
			if (orderbyMethodCall.Method.Name != "OrderBy")
				throw new NotSupportedException("Only OrderBy MethodCalls supported.");

			if (orderbyMethodCall.Parameters.Count != 2)
				throw new NotSupportedException("Only paramCount 2 supported.");

			if (orderbyMethodCall.Parameters[1].NodeType != ExpressionType.Lambda)
				throw new NotSupportedException("Only OrderBy parameter type Lambda supported.");

			LambdaExpression orderbyLambda = ((LambdaExpression) orderbyMethodCall.Parameters[1]);

			if (orderbyLambda.Body.NodeType != ExpressionType.MemberAccess)
				throw new NotSupportedException("Only MemberAccess expressions supported for lambda body.");

			MemberExpression memberExpression = (MemberExpression) orderbyLambda.Body;
			if (memberExpression.Member.DeclaringType.Name != "CountryInfo")
				throw new NotSupportedException("Only members of class CountryInfo supported.");
			if (memberExpression.Member.MemberType != MemberTypes.Property)
				throw new NotSupportedException("Only members of type Property supported.");
			switch (memberExpression.Member.Name) {
				case "Name":
					result.SortOrder = DataSortOrder.Name;
					break;
				case "Population":
					result.SortOrder = DataSortOrder.Population;
					break;
				default:
					throw new NotSupportedException("Only members named 'Name' or 'Population' supported.");
			}

			MethodCallExpression whereMethodCall = (MethodCallExpression) ((MethodCallExpression) expression).Parameters[0];
			if (whereMethodCall.Method.Name != "Where")
				throw new NotSupportedException("Only Where MethodCalls supported.");

			if (whereMethodCall.Parameters.Count != 2)
				throw new NotSupportedException("Only paramCount 2 supported.");

			if (whereMethodCall.Parameters[1].NodeType != ExpressionType.Lambda)
				throw new NotSupportedException("Only Where parameter type Lambda supported.");

			LambdaExpression whereLambda = ((LambdaExpression) whereMethodCall.Parameters[1]);

			if (whereLambda.Body.NodeType != ExpressionType.MethodCall && whereLambda.Body.NodeType != ExpressionType.MethodCallVirtual)
				throw new NotSupportedException("Only MethodCall(Virtual) expressions supported for lambda body.");

			MethodCallExpression lambdaMethodCall = (MethodCallExpression) whereLambda.Body;
			if (lambdaMethodCall.Method.Name != "StartsWith")
				throw new NotSupportedException("Only lambda body calls to string.StartsWith supported.");

			ConstantExpression filterPrefixExpression = (ConstantExpression) lambdaMethodCall.Parameters[0];
			if (filterPrefixExpression.Type != typeof(string))
				throw new NotSupportedException("Only string parameters to string.StartsWith supported.");

			result.FilterPrefix = (string) filterPrefixExpression.Value;

			return result;
		}
		#endregion

		#region Other stuff
		#region IQueryable<CountryInfo>.Execute<S>
		S IQueryable<CountryInfo>.Execute<S>(Expression expression) {
			throw new Exception("The method or operation is not implemented.");
		}
		#endregion

		#region IQueryable.CreateQuery
		IQueryable IQueryable.CreateQuery(Expression expression) {
			throw new Exception("The method or operation is not implemented.");
		}
		#endregion
		#region IQueryable.ElementType
		Type IQueryable.ElementType {
			get {
				throw new Exception("The method or operation is not implemented.");
			}
		}
		#endregion
		#region IQueryable.Execute
		object IQueryable.Execute(Expression expression) {
			throw new Exception("The method or operation is not implemented.");
		}
		#endregion

		#region IEnumerable.GetEnumerator

		IEnumerator IEnumerable.GetEnumerator( ) {
			// Parse the expression, execute the query and return
			// the list of results

			throw new Exception("The method or operation is not implemented.");
		}

		#endregion
		#endregion
	}

	public class SearchInfo {
		public SearchInfo( ) {

		}

		public SearchInfo(string filterPrefix, DataSortOrder sortOrder) {
			this.filterPrefix = filterPrefix;
			this.sortOrder = sortOrder;
		}

		private string filterPrefix;
		public string FilterPrefix {
			get {
				return filterPrefix;
			}
			set {
				filterPrefix = value;
			}
		}

		private DataSortOrder sortOrder;
		public DataSortOrder SortOrder {
			get {
				return sortOrder;
			}
			set {
				sortOrder = value;
			}
		}
	}
}
